#include "fraise.h"

int getHomography(IplImage *imgsrc, IplImage *imgdst, CvMat **H, int method) {
  /* Initialization block */
  CvMat *ptsrc, *ptdst;
  CvPoint2D32f* pts[2] = {0,0};
	/* Parameters required by cvCornerSubpix and cvPyrOpticalFlow */
	int count = MAX_COUNT, win_size = 10, level = 3, flags = 0;
  int icnt, kcnt;
  IplImage *pyrsrc, *pyrdst;
  char *status;
  count = MAX_COUNT;
  static int _sanity_count = 0;

  /* If we have reached a frame which is a multiple of TRIGGER_F2F_H
  Perform frame 2 frame homography, else we fall back to the opticalflow 
  based method which is reasonably fast */
  if (_sanity_count % TRIGGER_F2F_H == 0) {
    /* SURF or SIFT whichever is appropriate */
    fprintf(stdout, "F2F H Triggered : %d %d\n", _sanity_count, 
	    _sanity_count % TRIGGER_F2F_H);
    findHomographySURF (imgsrc, imgdst, H);
    /* Divide the entire homography matrix with the last element */
    cvConvertScale (*H, *H, 1./cvmGet(*H, 2, 2), 0);
    _sanity_count++;
    return 0;	
  }
  /* This would be performed most of the times */	
  pyrsrc = cvCreateImage( cvGetSize(imgsrc), 8, 1 );
  pyrdst = cvCreateImage( cvGetSize(imgdst), 8, 1 );
  switch (method) {
    case H_KLTF:
      pts[0] = kltf2pts (imgsrc, &count);
      break;
#ifdef HAVESIFT
    case H_SIFT:
      pts[0] = sift2pts (imgsrc, &count);
      break;
#endif
    case H_SURF:
      pts[0] = surf2pts (imgsrc, &count);
      break;
    case H_MSER:
      pts[0] = mser2pts (imgsrc, &count);
      break;
  }
  /*
  for (icnt=0; icnt<count; icnt++)
  fprintf (stderr, "%d:%3.4f, %3.4f\n", 
        		icnt, pts[0][icnt].x, pts[0][icnt].y);
  */
  cvFindCornerSubPix (imgsrc, pts[0], count,
        cvSize (win_size, win_size), cvSize(-1,-1),
        cvTermCriteria (CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 20, 0.03));
		
				
  /* Allocating memory for correspondening features */		
  if (count > MAX_COUNT) {
    status = (char*)cvAlloc (count);
    pts[1] = (CvPoint2D32f *) cvAlloc (count * sizeof(CvPoint2D32f));	
  } else {
    status = (char*)cvAlloc (MAX_COUNT);
    pts[1] = (CvPoint2D32f *) cvAlloc (MAX_COUNT * sizeof(CvPoint2D32f));	
  }
	
  cvCalcOpticalFlowPyrLK (imgsrc, imgdst, pyrsrc, pyrdst,
    pts[0], pts[1], count, cvSize(win_size, win_size), level, status, NULL,
    cvTermCriteria (CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 20, 0.03), flags);
  for (icnt=0, kcnt= 0; icnt <count; icnt++) {
    if (status[icnt]!=0)
      kcnt++;
  }
  ptsrc = cvCreateMat(kcnt, 2, CV_32FC1);
  ptdst = cvCreateMat(kcnt, 2, CV_32FC1);
  /* Converting the kpts to a Nx2 matrix usable by cvFindHomography */
  for (icnt=0, kcnt= 0; icnt <count; icnt++) {
    if (status[icnt]!=0) {
      cvmSet(ptsrc, kcnt, 0, pts[0][icnt].x);
      cvmSet(ptsrc, kcnt, 1, pts[0][icnt].y);

      cvmSet(ptdst, kcnt, 0, pts[1][icnt].x);
      cvmSet(ptdst, kcnt, 1, pts[1][icnt].y);
      kcnt++;
    }
  }

  *H = cvCreateMat (3, 3, CV_32FC1);
  cvFindHomography (ptsrc, ptdst, *H, CV_RANSAC, 5, NULL);
	/* Divide the entire homography matrix with the last element */
	cvConvertScale (*H, *H, 1./cvmGet(*H, 2, 2), 0);
		
	_sanity_count++;
	/* clean up */
  cvReleaseMat (&ptsrc);
  cvReleaseMat (&ptdst);
  cvReleaseImage (&pyrsrc);
  cvReleaseImage (&pyrdst);
  cvFree (&pts[0]);
  cvFree (&pts[1]); 
	
  return 0;
}
/* Function to return KLTF points corresponding to an image*/
CvPoint2D32f *kltf2pts (IplImage *imgsrc, int *count){
  IplImage *eig, *tmp;
  double quality, min_distance;
	CvPoint2D32f *kltpts;
				
	eig = cvCreateImage (cvGetSize(imgsrc), 32, 1);
  tmp = cvCreateImage (cvGetSize(imgsrc), 32, 1);
  quality = 0.01;
  min_distance = 10;
	kltpts = (CvPoint2D32f *) cvAlloc (*count * sizeof(CvPoint2D32f));
  cvGoodFeaturesToTrack (imgsrc, eig, tmp, kltpts, count,
                         quality, min_distance, 0, 3, 0, 0.04 );
	cvReleaseImage (&eig);
  cvReleaseImage (&tmp);
	return kltpts;
}
#ifdef HAVESIFT
/* Function to return sift points corresponding to an image that could be used 
 by the following subpixel routine, needs external library support */
CvPoint2D32f *sift2pts (IplImage *imgsrc, int *count){
  struct feature *siftfeat; 
  CvPoint2D32f *siftpts;
  int icnt;
		
  *count = sift_features (imgsrc, &siftfeat);
  if (*count > MAX_COUNT) {
    fprintf(stderr, "Increase MAX_COUNT to at least %d and recompile\n", 
	    *count);
    *count = MAX_COUNT;
  }
  siftpts = (CvPoint2D32f *) cvAlloc (*count * sizeof(CvPoint2D32f));
  for (icnt=0; icnt<*count; icnt++)
    siftpts[icnt] = cvPoint2D32f ((double) siftfeat[icnt].x, 
                     (double) siftfeat[icnt].y);
  return siftpts;
}
#endif
/* Function to return SURF points corresponding to an image that could be used 
 by the following subpixels routine */
CvPoint2D32f *surf2pts (IplImage *imgsrc, int *count){
  CvSURFParams sparams;
  CvSeq *surfkpts, *surfdesc;
  CvSeqReader surfreader;
  const CvSURFPoint *kpt;
  int icnt;
  CvPoint2D32f *surfpt;
				
  sparams = cvSURFParams (500, 1);
  cvExtractSURF (imgsrc, 0, &surfkpts, &surfdesc, __mmstg , sparams, 0);
  cvStartReadSeq (surfkpts, &surfreader, 0);
  *count = surfkpts->total;
  surfpt = (CvPoint2D32f *) cvAlloc (*count * sizeof(CvPoint2D32f));	
  for( icnt = 0; icnt < *count; icnt++ )  {
    kpt = (const CvSURFPoint*) surfreader.ptr;
    surfpt[icnt] = kpt->pt;
    CV_NEXT_SEQ_ELEM (surfreader.seq->elem_size, surfreader);
  }	
  return surfpt;		
}
/* Function to return MSER points corresponding to an image that could be used 
 by the following subpixels routine */
CvPoint2D32f *mser2pts (IplImage *imgsrc, int *count){
  int icnt, jcnt, kcnt;
  CvPoint *pt;
  CvMSERParams mparams;
  CvSeq *contours, *mserkpts;
  int estimate, ub, idx;
  CvPoint2D32f *mserpt;
  CvMat *t;
		
  mparams = cvMSERParams (5, 60, cvRound(.2*imgsrc->width*imgsrc->height), 
             .25f, .2, 200, 1.01, .003, 5);
  cvExtractMSER (imgsrc, NULL, &contours, __mmstg, mparams);
  /* Get an estimate of the total number of points */
  kcnt = 0;
  for (icnt=0; icnt <contours->total; icnt++){
    mserkpts = *(CvSeq**) cvGetSeqElem (contours, icnt);
		kcnt += mserkpts->total;
  }
  estimate = cvCeil(kcnt*FRAC+ contours->total);
  //fprintf(stderr, "%d Random features should be selected\n", estimate);
					
  /* Find random points from the ensemble of all points */
  mserpt = (CvPoint2D32f *) cvAlloc (estimate * sizeof(CvPoint2D32f));
			
  kcnt = 0;
  for (icnt=0; icnt <contours->total; icnt++){
    mserkpts = *(CvSeq**) cvGetSeqElem (contours, icnt);
    ub = mserkpts->total;
    t = randidxs (ub, FRAC);
    //fprintf(stderr, "Random features selected %d for contour %d\n", 
    //    t->rows, icnt);
				
    for (jcnt = 0; jcnt < t->rows; jcnt++){
      idx = (int) cvmGet(t, jcnt, 0);
      //fprintf(stderr, "index %d, %d: %d \n", jcnt, idx, kcnt);
						
      pt = CV_GET_SEQ_ELEM (CvPoint, mserkpts, idx); 
      //fprintf(stderr, "%d: %d, %d\n", count, pt->x, pt->y);	
        mserpt[kcnt] = cvPointTo32f(*pt);
      kcnt++;
    }
    cvReleaseMat (&t);
  }
  *count = kcnt;
  return mserpt;
}
/* Function to generate an array of random indices. ub being the upper bound 
 and frac is the fraction of total number of elements to be taken */
CvMat *randidxs(int ub, double frac) {
  int i;
  time_t secs;
  /* Change this to cvRNG(-1) for a pseudo random number generator */
  CvRNG rstate = cvRNG((int) (time(&secs)));
  CvMat* idxs = cvCreateMat(cvCeil(ub*frac), 1, CV_32FC1);
		
  cvRandArr(&rstate, idxs, CV_RAND_UNI, cvRealScalar(0), cvRealScalar(ub-1));
  for(i = 0; i <idxs->rows; i++ ){
    cvmSet(idxs, i, 0, cvCeil(cvmGet(idxs, i, 0)));	
//    fprintf(stderr, "Mat ELEM %d: %d\n", i, (int)cvmGet(idxs, i, 0));    
  }
  return idxs;
}
/* Function for computing frame to frame homography between source and 
 destination image pairs */
void findHomographySURF(IplImage *imgsrc, IplImage *imgdst, CvMat **H){
  CvSeq *srckpts, *srcdes, *dstkpts, *dstdes;
  CvMat *ptsrc, *ptdst;
  int icnt, n, nearest_neighbor;
  CvSeqReader reader, kreader;
  CvSeq *ptpairs;
  int *idx;
  const CvSURFPoint *kp;
  const float *descriptor;

  CvSURFParams params = cvSURFParams(500, 1);
  cvExtractSURF(imgsrc, 0, &srckpts, &srcdes, __mmstg , params, 0);
  cvExtractSURF(imgdst, 0, &dstkpts, &dstdes, __mmstg , params, 0);

  cvStartReadSeq(srckpts, &kreader, 0);
  cvStartReadSeq(srcdes, &reader, 0);

  ptpairs = cvCreateSeq (CV_32SC1, sizeof(CvSeq), sizeof(int), __mmstg );

  for( icnt = 0; icnt < srcdes->total; icnt++ )  {
    kp = (const CvSURFPoint*)kreader.ptr;
    descriptor = (const float*)reader.ptr;
    CV_NEXT_SEQ_ELEM ( kreader.seq->elem_size, kreader );
    CV_NEXT_SEQ_ELEM ( reader.seq->elem_size, reader );

    nearest_neighbor = naiveNearestNeighbor (descriptor, kp->laplacian, dstkpts, dstdes);
    if( nearest_neighbor >= 0 ){
      cvSeqPush(ptpairs, &icnt);
      cvSeqPush(ptpairs, &nearest_neighbor);
    }
  }

  n = (int)ptpairs->total/2;
  ptsrc = cvCreateMat (n, 2, CV_32FC1);
  ptdst = cvCreateMat (n, 2, CV_32FC1);

  for(icnt = 0; icnt < n; icnt++ ) {
    idx = (int *)CV_GET_SEQ_ELEM(int, ptpairs, icnt*2);
    cvmSet(ptsrc, icnt, 0, (double)((CvSURFPoint*)cvGetSeqElem(srckpts,*idx))->pt.x);
    cvmSet(ptsrc, icnt, 1, (double)((CvSURFPoint*)cvGetSeqElem(srckpts,*idx))->pt.y);

    idx = (int *)CV_GET_SEQ_ELEM(int, ptpairs, icnt*2+1);
    cvmSet(ptdst, icnt, 0, (double)((CvSURFPoint*)cvGetSeqElem(dstkpts,*idx))->pt.x);
    cvmSet(ptdst, icnt, 1, (double)((CvSURFPoint*)cvGetSeqElem(dstkpts,*idx))->pt.y);
  }

  *H = cvCreateMat (3, 3, CV_32FC1);
  cvFindHomography (ptsrc, ptdst, *H, CV_RANSAC, 5, NULL);

  cvReleaseMat (&ptsrc);
  cvReleaseMat (&ptdst);
}
/* helper functions for findHomographySURF */
int naiveNearestNeighbor( const float* vec, int laplacian,
                          const CvSeq* model_keypoints,
                          const CvSeq* model_descriptors ) {
  int length = (int)(model_descriptors->elem_size/sizeof(float));
  int i, neighbor = -1;
  double d, dist1 = 1e6, dist2 = 1e6;
  CvSeqReader reader, kreader;
  const CvSURFPoint* kp;
  const float* mvec;
  
  cvStartReadSeq( model_keypoints, &kreader, 0);
  cvStartReadSeq( model_descriptors, &reader, 0);

  for( i = 0; i < model_descriptors->total; i++ ) {
    kp = (const CvSURFPoint*)kreader.ptr;
    mvec = (const float*)reader.ptr;
    CV_NEXT_SEQ_ELEM( kreader.seq->elem_size, kreader );
    CV_NEXT_SEQ_ELEM( reader.seq->elem_size, reader );
    if( laplacian != kp->laplacian )
      continue;
    d = compareSURFDescriptors( vec, mvec, dist2, length );
    if( d < dist1 )
    {
      dist2 = dist1;
      dist1 = d;
      neighbor = i;
    }
    else if ( d < dist2 )
      dist2 = d;
  }
  if ( dist1 < 0.6*dist2 )
    return neighbor;
  return -1;
}
/* helper functions for findHomographySURF */
double compareSURFDescriptors( const float* d1, const float* d2, double best, int length ){
  double total_cost = 0;
  double t0, t1, t2, t3;
  int i;
  
  assert( length % 4 == 0 );
  for(i = 0; i < length; i += 4 )
  {
    t0 = d1[i] - d2[i];
    t1 = d1[i+1] - d2[i+1];
    t2 = d1[i+2] - d2[i+2];
    t3 = d1[i+3] - d2[i+3];
    total_cost += t0*t0 + t1*t1 + t2*t2 + t3*t3;
    if( total_cost > best )
      break;
  }
  return total_cost;
}
